#ifndef _MESH_H
#define _MESH_H

#include <iostream>
#include <vector>

#include "xvec.h"

class EdgeT;
class VertexT;

class DEdgeT {
  // will be using the fact that a pointer to a four byte structure
  // like an edge has zeroes in its two least significant bits
  enum { DIR_MASK = 0x00000001, PE_MASK = 0xFFFFFFFC };
public:
  DEdgeT() : m_p(0) {
  }
  
  DEdgeT(const DEdgeT& _de) : m_p(_de.m_p) {
  }

  DEdgeT(EdgeT* _e, int _dir) {
    m_p = (unsigned(_e) & PE_MASK) | (unsigned(_dir) & DIR_MASK);
  }
  
  EdgeT* Edge() const {
    return (EdgeT*)(m_p & PE_MASK);
  }
  
  int Dir() const {
    return (int)(m_p & DIR_MASK);
  }
  
  DEdgeT Flip() const {
    DEdgeT de(*this);
    de.m_p ^= DIR_MASK;
    return de;
  }

  DEdgeT Enext() const;
  
  void SetEnext(const DEdgeT& de) const;

  DEdgeT Enext2() const {
    return Enext().Enext();
  }

  bool Boundary() const {
    return !Enext();
  }

  bool AdjacentToBoundary() const {
    return Boundary() || Flip().Boundary();
  }

  VertexT* Org() const;
  VertexT* Dest() const;
  
  void SetOrg(VertexT* v) const;

  operator bool() const {
    return m_p!=0;
  }
  bool operator!() const {
    return m_p==0;
  }
  bool operator==(const DEdgeT& de) {
    return m_p==de.m_p;
  }
private:
  unsigned m_p;
};

inline std::ostream& operator<<(std::ostream& os, DEdgeT& de) {
  os << "(" << de.Edge() << ":" << de.Dir() << ")";
  return os;
}

class UmIt {
public:
  UmIt(const VertexT* v);
  DEdgeT Next() {
    if(!m_de)
      return m_de;
    
    //cerr << m_de << endl;

    DEdgeT der = m_de;
    m_de = m_de.Flip().Enext();

    if(m_de==m_de0)
      m_de = DEdgeT();
    
    return der;
  }
private:
  DEdgeT m_de, m_de0;
  bool m_bnd;
};

class VertexT {
public:
  VertexT() : m_color(0.6f, 0.6f, 0.6f) {}
  DEdgeT DEdge() const {
    return m_de;
  }
  DEdgeT& DEdge() {
    return m_de;
  }

  bool Boundary() const {
    return m_de.Boundary();
  }
  int Valence() const {
    int val = 0;
    UmIt ui(this);
    DEdgeT de;
    while(de=ui.Next()) {
      ++val;
    }
    return val;
  }

  const XVecf& Pos() const {
    return m_pos;
  }
  XVecf& Pos() {
    return m_pos;
  }

  const XVecf& Normal() const {
    return m_normal;
  }
  XVecf& Normal() {
    return m_normal;
  }

  const XVecf& Color() const {
    return m_color;
  }
  XVecf& Color() {
    return m_color;
  }

private:
  DEdgeT m_de;
  XVecf m_pos;
  XVecf m_normal;
  XVecf m_color;
};

UmIt::UmIt(const VertexT* v) : m_de(v->DEdge()), m_de0(v->DEdge()) {
  m_bnd = m_de.Boundary();
}

class EdgeT {
public:
  EdgeT() {
    m_vs[0] = m_vs[1] = 0;
  }
  DEdgeT DEdge(int _dir) const {
    return m_des[_dir];
  }
  
  DEdgeT& DEdge(int _dir) {
    return m_des[_dir];
  }

  VertexT* Vertex(int _dir) const {
    return m_vs[_dir];
  }

  VertexT*& Vertex(int _dir) {
    return m_vs[_dir];
  }
private:
  DEdgeT m_des[2];
  VertexT* m_vs[2];
};

DEdgeT DEdgeT::Enext() const {
  return (*this ? Edge()->DEdge(Dir()) : DEdgeT());
}

void DEdgeT::SetEnext(const DEdgeT& de) const {
  Edge()->DEdge(Dir()) = de;
}

VertexT* DEdgeT::Org() const {
  return (*this ? Edge()->Vertex(Dir()) : 0);
}

void DEdgeT::SetOrg(VertexT* v) const {
  Edge()->Vertex(Dir()) = v;
}

VertexT* DEdgeT::Dest() const {
  return (*this ? Edge()->Vertex(1-Dir()) : 0 );
}


class MeshT {
public:
  typedef std::vector<EdgeT*> EdgeCt;
  typedef std::vector<VertexT*> VertexCt;
public:
  MeshT() {}

  ~MeshT() {}

  const EdgeCt& Edges() const {
    return m_es;
  }
  const VertexCt& Verts() const {
    return m_vs;
  }

  // a simple procedure that properly assigns all the DEdges for vertices
  // even on the boundary
  void AssignVertEdges() {
    EdgeCt::const_iterator ei;
    for(ei=Edges().begin(); ei!=Edges().end(); ++ei) {
      (*ei)->Vertex(0)->DEdge() = DEdgeT(*ei, 0);
      (*ei)->Vertex(1)->DEdge() = DEdgeT(*ei, 1);
    }
    for(ei=Edges().begin(); ei!=Edges().end(); ++ei) {
      for(int dir=0; dir<2; ++dir) {
        DEdgeT de(*ei, dir);
        if(de.Boundary()) {
          de.Org()->DEdge() = de;
        }
      }
    }
  }

  static void SetNeighbors(DEdgeT& de, DEdgeT& den) {
    if(de)
      de.SetEnext(den);
  }

  static void AssignVertexDEdge(const DEdgeT& des) {
    if(!des)
      return;
    // go in the backward direction till hit itself or the boundary
    DEdgeT de = des;
    while(de) {
      if(de.Boundary()) {
        assert(de.Org()==des.Org());
        de.Org()->DEdge() = de;
        return;
      }
      de = de.Enext2().Flip();
      if(de==des) {
        assert(de.Org()==des.Org());
        de.Org()->DEdge() = de;
        return;
      }
    }
    // then assign
  }

  void AddVertex(VertexT* v) {
    m_vs.push_back(v);
  }

  void AddEdge(EdgeT* e) {
    m_es.push_back(e);
  }

private:
  EdgeCt m_es;
  VertexCt m_vs;
};

class MeshTSoupT : public ITriSoup {
public:
  MeshTSoupT(const MeshT& _mesh) : m_smesh(_mesh) {
    m_bmin = XVecf(1e+20,  1e+20,  1e+20);
    m_bmax = XVecf(-1e+20, -1e+20, -1e+20);

    MeshT::VertexCt::const_iterator vi;
    for(vi=m_smesh.Verts().begin(); vi!=m_smesh.Verts().end(); ++vi) {
      m_bmin = m_bmin.Min((*vi)->Pos());
      m_bmax = m_bmax.Max((*vi)->Pos());
    }
  }

  // get three vertex indices
  virtual void ResetTriangles() {
    m_ei = m_smesh.Edges().begin();
    m_dir = 0;
  }

  // get three vertex positions
  virtual bool GetNextTriangle(XVecf* pcoords, XVecf* pnormals=0, XVecf* pcolors=0) {
    if(m_ei==m_smesh.Edges().end())
      return false;
    // find a edge that is of minimal index in the triangle and then draw
    while(!minimal_edge()) {
      advance();
      if(m_ei==m_smesh.Edges().end())
        return false;
    }
    // fill the coordinates
    DEdgeT de(*m_ei, m_dir);
    DEdgeT den = de.Enext();

    pcoords[0] = de.Org()->Pos();
    pcoords[1] = de.Dest()->Pos();
    pcoords[2] = den.Dest()->Pos();

    if(pnormals) {
      pnormals[0] = de.Org()->Normal();
      pnormals[1] = de.Dest()->Normal();
      pnormals[2] = den.Dest()->Normal();
    }

    if(pcolors) {
      pcolors[0] = de.Org()->Color();
      pcolors[1] = de.Dest()->Color();
      pcolors[2] = den.Dest()->Color();
    }

    advance();

    return true;
  }

  // bbox
  virtual XVecf MinCoord() {
    return m_bmin;
  }

  virtual XVecf MaxCoord() {
    return m_bmax;
  }

protected:
  void advance() {
    m_dir = 1 - m_dir;
    if(m_dir==0)
      ++m_ei;
  }
  bool minimal_edge() {
    DEdgeT de(*m_ei, m_dir);
    DEdgeT den = de.Enext();
    if(!den)
      return false;
    DEdgeT denn = den.Enext();
    return (de.Edge()<den.Edge()) && (de.Edge()<denn.Edge());
  }

private:
  const MeshT& m_smesh;
  
  MeshT::EdgeCt::const_iterator m_ei;
  int m_dir;

  XVecf m_bmax, m_bmin;
};



#endif