/************************************** 

 Simple geometry viewer

 Left mouse:    rotate the scene
 Shift+Left mouse:    pan the scene
 Middle mouse:  zoom
 Right mouse:   menu
 Keyboard:      press ESC to quit, SPACE to reset view

 The function InitGeometry() initializes
 the geometry that will be displayed.

***************************************/

#include <assert.h>
#include <math.h>
#include <stdlib.h>
#include <GL/glut.h>

#include <iostream>

#include "xvec.h"
#include "tsoup.h"
#include "mesh.h"

using namespace std;

const float M_PI = 3.1415926f;

// GLOBALS

/* Viewer state */

float sphi=90.0, stheta=45.0;
float sdepth = 10;
float zNear=1.0, zFar=100.0;
float aspect = 5.0/4.0;
float xcam = 0, ycam = 0;
long xsize, ysize;
int downX, downY;
bool leftButton = false, middleButton = false;
int i,j;
GLfloat light0Position[] = { 0, 1, 0, 1.0}; 
int displayMenu, mainMenu;
enum {WIREFRAME, HIDDENLINE, FLATSHADED, SMOOTHSHADED, COLORSMOOTHSHADED};
int displayMode = WIREFRAME;
int modShift = 0;
float xpan = 0.0f, ypan = 0.0f;
float g_fov=45.0f;
XVecf g_center;
float g_sdepth;

ITriSoup* g_pts = 0; // points to the current triangular soup
MeshT m; // mesh

////////////////////////////////

void SetBBox(const XVecf& bmin, const XVecf& bmax);

void InitGeometry() {
  const int NV = 4;
  const int NE = 6;
  
  EdgeT* es[NE];
  VertexT* vs[NV];

  int i;
  for(i=0; i<NV; ++i) {
    vs[i] = new VertexT;
    m.AddVertex(vs[i]);
  }

  for(i=0; i<NE; ++i) {
    es[i] = new EdgeT;
    m.AddEdge(es[i]);
  }

  // fill in coordinates
  vs[0]->Pos() = XVecf(0.0f, 0.0f, 0.0f);
  vs[1]->Pos() = XVecf(10.0f, 0.0f, 0.0f);
  vs[2]->Pos() = XVecf(0.0f, 10.0f, 0.0f);
  vs[3]->Pos() = XVecf(0.0f, 0.0f, 10.0f);
  // and normals
  vs[0]->Normal() = XVecf(-0.577f, -0.577f, -0.577f);
  vs[1]->Normal() = XVecf(0.0f, -0.7f, -0.7f);
  vs[2]->Normal() = XVecf(-0.7f, 0.0f, -0.7f);
  vs[3]->Normal() = XVecf(-0.7f, -0.7f, 0.0f);
  // and colors
  vs[0]->Color() = XVecf(0.5f, 0.5f, 0.5f); // gray
  vs[1]->Color() = XVecf(0.2f, 0.2f, 1.0f); // blue
  vs[2]->Color() = XVecf(0.2f, 1.0f, 0.2f); // green
  vs[3]->Color() = XVecf(1.0f, 0.2f, 0.2f); // red

  // set up edge vertices 
  es[0]->Vertex(0) = vs[0];
  es[0]->Vertex(1) = vs[1];

  es[1]->Vertex(0) = vs[0];
  es[1]->Vertex(1) = vs[2];

  es[2]->Vertex(0) = vs[0];
  es[2]->Vertex(1) = vs[3];

  es[3]->Vertex(0) = vs[1];
  es[3]->Vertex(1) = vs[2];

  es[4]->Vertex(0) = vs[2];
  es[4]->Vertex(1) = vs[3];

  es[5]->Vertex(0) = vs[3];
  es[5]->Vertex(1) = vs[1];

  // set up enext relations among edges -- as many as there are corners
  // that is nine
  MeshT::SetNeighbors( DEdgeT(es[0], 0), DEdgeT(es[5], 1) );
  MeshT::SetNeighbors( DEdgeT(es[5], 1), DEdgeT(es[2], 1) );
  MeshT::SetNeighbors( DEdgeT(es[2], 1), DEdgeT(es[0], 0) );

  MeshT::SetNeighbors( DEdgeT(es[1], 0), DEdgeT(es[3], 1) );
  MeshT::SetNeighbors( DEdgeT(es[3], 1), DEdgeT(es[0], 1) );
  MeshT::SetNeighbors( DEdgeT(es[0], 1), DEdgeT(es[1], 0) );

  MeshT::SetNeighbors( DEdgeT(es[2], 0), DEdgeT(es[4], 1) );
  MeshT::SetNeighbors( DEdgeT(es[4], 1), DEdgeT(es[1], 1) );
  MeshT::SetNeighbors( DEdgeT(es[1], 1), DEdgeT(es[2], 0) );

  m.AssignVertEdges();

  // create triangular soup interface for the viewer
  g_pts = new MeshTSoupT(m);
  SetBBox(g_pts->MinCoord(), g_pts->MaxCoord());
}

void MyIdleFunc(void) {

  // things to do while idle

  glutPostRedisplay();
}


void RunIdleFunc(void) {

  glutIdleFunc(MyIdleFunc);
}


void PauseIdleFunc(void) {

  glutIdleFunc(NULL);
}


void DrawSmoothShaded(void) {

  glColor3f(0.8f, 0.2f, 0.8f);

  glBegin ( GL_TRIANGLES ) ;

  XVecf xyzs[3];
  XVecf normals[3];

  g_pts->ResetTriangles();

  while(g_pts->GetNextTriangle(xyzs, normals)) {
    glNormal3fv( (float*)normals[0] );
    glVertex3fv( (float*)xyzs[0]  );

    glNormal3fv( (float*)normals[1] );
    glVertex3fv( (float*)xyzs[1] );

    glNormal3fv( (float*)normals[2] );
    glVertex3fv( (float*)xyzs[2] );
  }
  
  glEnd ( ) ;
}

void DrawColorSmoothShaded(void) {

  glBegin ( GL_TRIANGLES ) ;

  XVecf xyzs[3];
  XVecf normals[3];
  XVecf colors[3];

  g_pts->ResetTriangles();

  while(g_pts->GetNextTriangle(xyzs, normals, colors)) {
    glColor3fv( (float*)colors[0] );
    glNormal3fv( (float*)normals[0] );
    glVertex3fv( (float*)xyzs[0]  );

    glColor3fv( (float*)colors[1] );
    glNormal3fv( (float*)normals[1] );
    glVertex3fv( (float*)xyzs[1] );

    glColor3fv( (float*)colors[2] );
    glNormal3fv( (float*)normals[2] );
    glVertex3fv( (float*)xyzs[2] );
  }
  
  glEnd ( ) ;
}


void DrawWireframe(void) {

  glColor3f(1.0, 1.0, 1.0);

  XVecf xyzs[3];
  XVecf normals[3];

  g_pts->ResetTriangles();

  while(g_pts->GetNextTriangle(xyzs, normals)) {
    glBegin(GL_LINE_STRIP);
    glVertex3fv( (float*)xyzs[0]  );
    glVertex3fv( (float*)xyzs[1] );
    glVertex3fv( (float*)xyzs[2] );
    glVertex3fv( (float*)xyzs[0]  );
    glEnd();
  }
}


void DrawFlatShaded(void) {

  glEnable(GL_POLYGON_OFFSET_FILL);
  glColor3f(0.8f, 0.2f, 0.8f);

  glBegin ( GL_TRIANGLES ) ;
  XVecf xyzs[3];
  XVecf normals[3];

  g_pts->ResetTriangles();

  while(g_pts->GetNextTriangle(xyzs, normals)) {
    XVecf n = (xyzs[1]-xyzs[0]).cross(xyzs[2]-xyzs[0]);
    float nd = n.l2();
    n /= nd;

    glNormal3fv( (float*)n );

    glVertex3fv( (float*)xyzs[0] );
    glVertex3fv( (float*)xyzs[1] );
    glVertex3fv( (float*)xyzs[2] );

  }

  glEnd ( ) ;

  glDisable(GL_POLYGON_OFFSET_FILL);
}


void DrawHiddenLine(void) {

  glEnable(GL_POLYGON_OFFSET_FILL);
  glColor3f(0,0,0);

  glBegin ( GL_TRIANGLES ) ;

  XVecf xyzs[3];
  XVecf normals[3];

  g_pts->ResetTriangles();
  while(g_pts->GetNextTriangle(xyzs, normals)) {
    glVertex3fv( (float*)xyzs[0] );
    glVertex3fv( (float*)xyzs[1] );
    glVertex3fv( (float*)xyzs[2] );
  }

  glEnd ( ) ;

  glDisable(GL_POLYGON_OFFSET_FILL);
    
  glColor3f(1.0,1.0,1.0);
  glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);

  glBegin ( GL_TRIANGLES ) ;

  g_pts->ResetTriangles();
  while(g_pts->GetNextTriangle(xyzs, normals)) {
    glVertex3fv( (float*)xyzs[0] );
    glVertex3fv( (float*)xyzs[1] );
    glVertex3fv( (float*)xyzs[2] );
  }

  glEnd ( ) ;

  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
}


void ReshapeCallback(int width, int height) {

  xsize = width; 
  ysize = height;
  aspect = (float)xsize/(float)ysize;
  glViewport(0, 0, xsize, ysize);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  glutPostRedisplay();
}


void SetDisplayMenu(int value) {
 
  displayMode = value;
  switch(value) {

    case WIREFRAME   : 
      glShadeModel(GL_FLAT); 
      glDisable(GL_LIGHTING);
      break;

    case HIDDENLINE: 
      glShadeModel(GL_FLAT); 
      glDisable(GL_LIGHTING);
      break;

    case FLATSHADED  : 
      glShadeModel(GL_FLAT); 
      glEnable(GL_LIGHTING);
      break;

    case SMOOTHSHADED: 
      glShadeModel(GL_SMOOTH); 
      glEnable(GL_LIGHTING);
      break;

    case COLORSMOOTHSHADED:
      glShadeModel(GL_SMOOTH); 
      glEnable(GL_LIGHTING);
      break;
  }

  glutPostRedisplay();
}


void SetMainMenu(int value) {

  switch(value) {

    case 99:    exit(0);   break;
  }
}

void SetBBox(const XVecf& bmin, const XVecf& bmax) {
  g_center = 0.5f*(bmin+bmax);
  float radius = bmax.dist( bmin );

  zNear = .2f * radius / sinf(0.5f*g_fov*M_PI/180.0f);
  zFar = zNear + 2.0f * radius;
  g_sdepth = ( zNear + radius );
  zNear *= .1f;
  zFar *= 10;

  sdepth = g_sdepth;
}

void DisplayCallback(void) {

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(g_fov, aspect, zNear, zFar);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity(); 

  glTranslatef(xpan, ypan, -sdepth);
  glRotatef(-stheta, 1.0, 0.0, 0.0);
  glRotatef(sphi, 0.0, 0.0, 1.0);

  //cerr << g_center << endl;
  glTranslatef(-g_center.x(), -g_center.y(), -g_center.z());

  if(g_pts) {
    switch (displayMode) {

      case WIREFRAME: DrawWireframe(); 
      break;

      case HIDDENLINE: DrawHiddenLine();
      break;

      case FLATSHADED: DrawFlatShaded();
      break;

      case SMOOTHSHADED: DrawSmoothShaded();
      break;

      case COLORSMOOTHSHADED: DrawColorSmoothShaded();
      break;
    }
  }
  glutSwapBuffers();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
}


void KeyboardCallback(unsigned char ch, int x, int y) {

  switch (ch) {
  case 32: 
    sphi=90.0;
    stheta=45.0;
    sdepth = g_sdepth;
    xpan = 0.0f;
    ypan = 0.0f;
    break;
  case 27: 
    exit(0);
    break;
  }

  glutPostRedisplay();
}


void MouseCallback(int button, int state, int x, int y) {

  downX = x;
  downY = y;
  leftButton = ((button == GLUT_LEFT_BUTTON) && 
		(state == GLUT_DOWN));
  middleButton = ((button == GLUT_MIDDLE_BUTTON) && 
		  (state == GLUT_DOWN));
  modShift = (glutGetModifiers() & GLUT_ACTIVE_SHIFT);

  glutPostRedisplay();
}


void MotionCallback(int x, int y) {


  if (leftButton) { 
    if(!modShift) { // rotate
      sphi += (float)(x - downX) / 4.0;
      stheta += (float)(downY - y) / 4.0;
    } else { // pan
      xpan += (float)(x - downX)*sdepth/zNear/xsize;
      ypan += (float)(downY - y)*sdepth/zNear/ysize;
    }
  }

  if (middleButton) { // scale
    sdepth += (float)(downY - y) / 10.0;
  }

  downX = x;
  downY = y;
  glutPostRedisplay();
}


void InitGL() {

  glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(500, 500);
  glutCreateWindow("eecs598-1 TSoup Viewer");
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LEQUAL);
  glClearColor(0.0, 0.0, 0.0, 0.0);
  glPolygonOffset(1.0, 1.0);
  glDisable(GL_CULL_FACE);
  glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
  glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glEnable(GL_COLOR_MATERIAL);
  glColorMaterial(GL_FRONT, GL_DIFFUSE);
  glLightfv (GL_LIGHT0, GL_POSITION, light0Position);
  glEnable(GL_LIGHT0);

  glutReshapeFunc(ReshapeCallback);
  glutDisplayFunc(DisplayCallback);

  glutKeyboardFunc(KeyboardCallback);
  glutMouseFunc(MouseCallback);
  glutMotionFunc(MotionCallback);
}


void InitMenu() {

  displayMenu = glutCreateMenu(SetDisplayMenu);
  glutAddMenuEntry("Wireframe", WIREFRAME);
  glutAddMenuEntry("Hidden Line", HIDDENLINE);
  glutAddMenuEntry("Flat Shaded", FLATSHADED);
  glutAddMenuEntry("Smooth Shaded", SMOOTHSHADED);
  glutAddMenuEntry("Color Smooth Shaded", COLORSMOOTHSHADED);

  mainMenu = glutCreateMenu(SetMainMenu);
  glutAddSubMenu("Display", displayMenu);
  glutAddMenuEntry("Exit", 99);
  glutAttachMenu(GLUT_RIGHT_BUTTON);
}


void main(int argc, char **argv) {

  glutInit(&argc, argv);

  InitGL();

  InitMenu();

  InitGeometry();

  glutMainLoop();
}
